/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sstream>
#include <limits>
#include <iostream>

#include "libtkcommon/debug.h"
#include "algorithm.h"

using namespace Tk::Algorithm;

const OffsetDetails & 
Offsets:: operator[] (const Offset &offset) const
{
  for (Offsets::const_iterator iter( this->begin() );
       iter != this->end();
       ++iter)
    if (iter->offset == offset)
      return *iter;
  
  throw std::runtime_error(std::string("Offset not found"));
}

Algorithm:: Algorithm (const String &spec)
  : _outcome_count(0)
{
  //WDEBUG_OUT << L"Algorithm spec: '" << spec << "'" << std::endl;
  
  // extract weights
  SmallNumbers weights;
  std::wistringstream strm(spec);
  unsigned i;
  while (strm >> i)
    weights.push_back (i);
  
  //
  // repeat the first half of the list and add it to the end
  //
  SmallNumbers reverse_list;
  if (weights.size() > 1)
    {
      //                                         vv skip the middle weight
      for (SmallNumbers::reverse_iterator iter = ++(weights.rbegin());
	   iter != weights.rend();
	   ++iter)
	reverse_list.push_back(*iter);
      
      for (SmallNumbers::const_iterator iter = reverse_list.begin();
	   iter != reverse_list.end();
	   ++iter)
	weights.push_back(*iter);
    }

  
  SmallNumbers::const_iterator weight;

  //
  // offsets
  //
  Offset offset( -(( weights.size() - 1 ) / 2) );
  for (weight = weights.begin();
       weight != weights.end();
       ++weight, ++offset)
    if (*weight > 0)
      _offsets.push_back( OffsetDetails(offset, *weight) );
  
  //
  // outcomes
  //
  Offsets weight_counts(_offsets);
  bool set_some;
  do
    {
      set_some = false;
      for (Offsets::iterator count_iter( weight_counts.begin() );
	   count_iter != weight_counts.end();
	   ++count_iter)
	{
	  if (count_iter->weight > 0)
	    {
	      _outcome_pattern.push_back(count_iter->offset);
	      --(count_iter->weight);
	      set_some = true;
	    }
	}
    }
  while (set_some);

  _outcome_count = _outcome_pattern.size();


  //
  // probabilities
  //
  const Probability outcome_count_d(_outcome_count);
  for (Offsets::iterator offset_iter( _offsets.begin() );
       offset_iter != _offsets.end();
       ++offset_iter)
    offset_iter->probability = static_cast<Probability>(offset_iter->weight) / outcome_count_d;
  

  // some nice debugging output
  /*
  std::wostringstream outcomes;
  for (i = 0; i < _outcome_count; ++i)
    outcomes << L"[" << _outcome_pattern[i %_outcome_pattern.size()] << L"]";

  std::wostringstream probabilities;
  for (OffsetProbabilities::const_iterator iter = _probabilities.begin();
       iter != _probabilities.end();
       ++iter)
    probabilities << L" P(" << iter->first << L")=" << iter->second;

  WDEBUG_OUT << L"Algorithm"
	     << L"; outcome count: " << _outcome_count
	     << L"; bit descriptor: " << _bit_desc
    	     << L"; outcomes: " << outcomes.str()
    	     << L"; probabilities:" << probabilities.str()
	     << std::endl;
  */
}


Offset
Algorithm:: choose_offset(unsigned choice) const
{
  return _outcome_pattern[choice % _outcome_pattern.size()];
}

const Offsets &
Algorithm:: get_offsets() const
{
  return _offsets;
}

const Tk::SmallNumbers &
Algorithm:: get_outcome_pattern() const
{
  return _outcome_pattern;
}

unsigned
Algorithm:: get_outcome_count() const
{
  return _outcome_count;
}
