/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <limits>
#include <iomanip>
#include <cmath>

#ifdef _OPENMP
#include <omp.h>
#endif

#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_hyperg.h>
#include <gsl/gsl_sf_result.h>
#include <gsl/gsl_sf_gamma.h>

#include "libtkcommon/config.h"
#include "libtkcommon/intl.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/numerical-types.h"
#include "libtkquery/parse.h"
#include "libtkdb/exception.h"
#include "analyser.h"

/* TK_OPENMP_PROFILE
 * 
 * Enable profiling of thread usage
 */
//#define TK_OPENMP_PROFILE

/* TK_POT_PROGRESS
 * 
 * With this #defined, the progress of the POT calculation will be
 * displayed on std::wcout.  This adds a slight overhead to the
 * calculation routine (about 15 seconds onto of a 13 minute run on a
 * 3.0GHz Intel Core2 Duo.  This is only 3% but that could be
 * significant in wall-clock time for large data sets.)
 */
//#define TK_POT_PROGRESS


using namespace Tk::Analysis;

const Tk::Rational*
PositionFunction:: operator()(const Position &pos) const
{
  PositionFunction::const_iterator pos_iter( find(pos) );
  if (pos_iter == end())
    return 0;
  
  return &pos_iter->second;
}

Tk::Rational
PositionFunction:: calc_rss(const Probability &prob) const
{
  Tk::Rational difference;
  Tk::Rational square;
  Tk::Rational SS_E(0);
  for (PositionFunction::const_iterator pos_iter( begin() );
       pos_iter != end();
       ++pos_iter)
    {
      difference = pos_iter->second - prob;
      square = difference * difference;
      SS_E += square;
    }
  
  return SS_E;
}

Tk::Rational
PositionFunction:: calc_mean() const
{
  Tk::Rational sum(0);
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    sum += iter->second;
  
  return sum / static_cast<Tk::Rational>( size() );
}

Tk::Rational
PositionFunction:: calc_variance(const Tk::Rational &mean) const
{
  Tk::Rational sum(0);
  Tk::Rational difference;
  Tk::Rational square;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      square = difference * difference;
      sum += square;
    }
  return sum / static_cast<Tk::Rational>( size() );
}

Tk::Real
PositionFunction:: calc_skewness(const Tk::Rational &mean, const Tk::Rational &variance) const
{
  Tk::Rational sum(0);
  Tk::Rational difference;
  Tk::Rational cube;
  
  if (size() == 1)
    return 0.0;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      cube = difference * difference * difference;
      sum += cube;
    }
  
  return ( sqrt( static_cast<Tk::Real>(size()) ) * static_cast<Tk::Real>(sum) ) / static_cast<Tk::Real>(pow(variance.get_d(), 2.0/3.0));
}

OffsetStatisticsSet:: OffsetStatisticsSet(const Tk::Algorithm::Offsets &offsets,
					  const PositionOffsetDistributions &offset_distributions,
					  const PositionCounts &position_counts)
  : _position_totals(position_counts)
{
  /* Offset list and theoretical probabilities */
  for (Tk::Algorithm::Offsets::const_iterator prob_iter( offsets.begin() );
       prob_iter != offsets.end();
       ++prob_iter)
    {
      _offsets.push_back(prob_iter->offset);
      _offset_stats[prob_iter->offset]._probability = prob_iter->probability;
    }
  
  /* Distributions and counts*/
  OffsetDistribution offset_position_counts;
  for (PositionCounts::const_iterator pos_iter( position_counts.begin() );
       pos_iter != position_counts.end();
       ++pos_iter)
    {
      const OffsetDistribution &offset_distribution( offset_distributions.find(pos_iter->first)->second );
      for (OffsetDistribution::const_iterator offset_iter( offset_distribution.begin() );
	   offset_iter != offset_distribution.end();
	   ++offset_iter)
	{
	  /* Distributions */
	  Tk::Rational div(static_cast<Tk::Rational>(offset_iter->second) / pos_iter->second);
	  PositionFunction::value_type pos_func_row(pos_iter->first, div);
	  _offset_stats[offset_iter->first]._position_function.insert(pos_func_row);
	  
	  /* Counts */
	  _offset_stats[offset_iter->first]._position_counts[pos_iter->first] = offset_iter->second;
	  Count &position_count_max( _offset_stats[offset_iter->first]._position_count_max );
	  if (offset_iter->second > position_count_max)
	    position_count_max = offset_iter->second;
	}
    }
  
  /* RSS, mean and variance */
  for (OffsetStatisticsMap::iterator offset_iter( _offset_stats.begin() );
       offset_iter != _offset_stats.end();
       ++offset_iter)
    {
      const PositionFunction &pos_func( offset_iter->second._position_function );
      if (pos_func.size() < 1)
	continue;
      
      OffsetStatistics &stats( offset_iter->second );
      stats._mean = pos_func.calc_mean();
      stats._rss = pos_func.calc_rss(stats._probability);
      stats._variance = pos_func.calc_variance(stats._mean);
      stats._skewness = pos_func.calc_skewness(stats._mean, stats._variance);
    }
}

/*
static void
spin(std::wostream &stream, bool finish = false)
{
  static const std::wstring seq(L"|/-\\|/-\\");
  static int spin_index(-1);
  
  if (finish)
    {
      spin_index = -1;
      stream << L'\b' << std::flush;
      return;
    }
  
  if (spin_index == -1)
    spin_index = 0;
  else
    stream << L'\b';
    
  ++spin_index;
  if (spin_index == seq.size())
    spin_index = 0;
  
  stream << seq[spin_index] << std::flush;
}
*/

#ifdef TK_POT_PROGRESS

typedef uint64_t HardBigInt;

static void
progress(std::wostream &stream, const HardBigInt &i, const HardBigInt &n)
{
  static int char_count(-1);
  
  for (; char_count > 0; --char_count)
    stream << L'\b';
  
  std::wostringstream str;
  HardReal perc( ((static_cast<HardReal>(i) / n) * 100) );
  str << i
      << L'/'
      << n
      << L" ("
      << std::fixed
      << std::setprecision(1)
      << perc
      << L"%)";
  
  stream << str.str() << std::flush;
  
  if (i >= n)
    char_count == -1;
  else
    char_count = str.str().size();
}
#endif /* TK_POT_PROGRESS */

HardReal
Statistics:: Pot(register HardReal r, const Slots &event_slots, ThreadContextMap &thread_profile)
{
  register HardReal Pot(0.0);
#ifdef TK_POT_PROGRESS
  const HardBigInt s( event_slots.size() - 1 );
  const HardBigInt n( (s*(s + 1))/2 );
  register HardBigInt i(0);
#endif
  
#ifdef _OPENMP
  omp_set_nested(1);

#  ifdef TK_POT_PROGRESS

#    define TK_POT_REDUCTION reduction(+:ii)

#    if _OPENMP < 200805
#      ifdef TK_OPENMP_PROFILE
#        define TK_OPENMP_SHARED thread_profile,i,
#      else
#        define TK_OPENMP_SHARED i,
#      endif
#    else
#      ifdef TK_OPENMP_PROFILE
#        define TK_OPENMP_SHARED thread_profile,event_slots,i,
#      else
#        define TK_OPENMP_SHARED event_slots,i,
#      endif
#    endif
#  else
#    if _OPENMP < 200805
#      ifdef TK_OPENMP_PROFILE
#        define TK_OPENMP_SHARED thread_profile,
#      else
#        define TK_OPENMP_SHARED 
#      endif
#    else
#      ifdef TK_OPENMP_PROFILE
#        define TK_OPENMP_SHARED thread_profile,event_slots,
#      else
#        define TK_OPENMP_SHARED event_slots,
#      endif
#    endif

#  endif

#  pragma omp parallel for schedule(guided) default(none) shared(TK_OPENMP_SHARED r, std::wcout) reduction(+:Pot)
#endif

#if defined(_OPENMP) && _OPENMP < 200805
  for(long J = 0; J < event_slots.size(); ++J)
#else
  for(Slots::const_iterator J = event_slots.begin(); J < event_slots.end(); ++J)
#endif
    {
#ifdef TK_POT_PROGRESS
      register HardBigInt ii(0);
#endif
      
#ifdef _OPENMP
# ifdef TK_POT_PROGRESS
#  pragma omp parallel for default(none) shared(TK_OPENMP_SHARED r, std::wcout, J) reduction(+:Pot) TK_POT_REDUCTION
#  else
#  pragma omp parallel for default(none) shared(TK_OPENMP_SHARED r, std::wcout, J) reduction(+:Pot)
# endif
#endif

#if defined(_OPENMP) && _OPENMP < 200805
      for (long I = 0; I < J; ++I)
#else
      for (Slots::const_iterator I = event_slots.begin(); I < J; ++I)
#endif
	{
	  HardInteger nI, nJ;
#if defined(_OPENMP) && _OPENMP < 200805
	  nI = event_slots[I];
	  nJ = event_slots[J];
#else
	  nI = *I;
	  nJ = *J;
#endif
	  
	  /***************************/
	  Pot += pow(r, abs(nI - nJ));
	  /***************************/

	  
#ifdef TK_POT_PROGRESS
	  ++ii;
#endif
	  
#ifdef TK_OPENMP_PROFILE
# ifdef _OPENMP
#  pragma omp critical
# endif
	  ++thread_profile[2][omp_get_thread_num()];
#endif
	}

#ifdef _OPENMP
# pragma omp critical
      {
#endif

#ifdef TK_POT_PROGRESS
	i += ii;
	progress(std::wcout, i, n);
#endif

#ifdef TK_OPENMP_PROFILE
	++thread_profile[1][omp_get_thread_num()];
#endif

#ifdef _OPENMP
      }
#endif
    }
  
  return Pot;
}

HardReal
Statistics:: Pot_expectation(HardInteger N, HardInteger M, HardReal r)
{
  HardReal term_1( (static_cast<HardReal>(N)*(N-1)) / (static_cast<HardReal>(2)*M*(M-1)) );
  HardReal term_2( S(r, M) );
  return term_1 * term_2;
  
  /*
  HardReal term_1( (static_cast<HardReal>(N)*(N-1)) / (static_cast<HardReal>(M)*(M-1)) );
  HardReal term_2( r / (1-r) );
  HardReal term_3( M - (1 / (1-r)) );
  return term_1 * term_2 * term_3;
  */
}

HardReal
Statistics:: S(HardReal r, HardInteger M)
{
  HardReal term_1( (2*r)/(1-r) );
  HardReal term_2( static_cast<HardReal>(M) - 1 - (r/(1-r))*( 1 - pow(r, M-1) ) );
  return term_1 * term_2;
  //  return ( (2*r)/(1-r) ) * ( M - (1/(1-r)) );
}

HardReal
Statistics:: G(HardReal r, HardInteger M)
{
  return pow(r/(1-r), 2) * ( 4*static_cast<HardReal>(M) + 2/(1-(r*r)) - 8/(1-r) );
}

HardReal
Statistics:: Pot_sd(HardInteger N, HardInteger M, HardReal r)
{
  HardReal NM_ratio( (static_cast<HardReal>(N)*(N-1)*(N-2)*(N-3)) / (static_cast<HardReal>(M)*(M-1)*(M-2)*(M-3)) );
  HardReal S_r2( S(pow(r, 2), M) );
  HardReal G_r( G(r, M) );
  
  HardReal term_1ab( (static_cast<HardReal>(N)*N*(N-1)*(N-1)) / (static_cast<HardReal>(M)*M*(M-1)*(M-1)) );
  HardReal term_1a( NM_ratio - term_1ab );
  HardReal term_1b( pow(S(r, M), 2) );
  HardReal term_1( term_1a * term_1b );
  
  HardReal term_2c( S_r2 - 2*G_r );
  HardReal term_2( 2 * NM_ratio * term_2c );
  
  HardReal term_3b( (static_cast<HardReal>(N)*(N-1)*(N-2)) / (static_cast<HardReal>(M)*(M-1)*(M-2)) );
  HardReal term_3c( G_r - S_r2 );
  HardReal term_3( 4 * term_3b * term_3c );
  
  HardReal term_4b( (static_cast<HardReal>(N)*(N-1)) / (static_cast<HardReal>(M)*(M-1)) );
  HardReal term_4( 2 * term_4b * S_r2 );
  
  return sqrt((term_1 + term_2 + term_3 + term_4) / 4);
}


Statistics:: Statistics(Db::Database &database,
			const IntentCode &intent_code,
			const Tk::Algorithm::Offsets &offsets,
			const Db::Ids &experiments,
			HardInteger S)
  : _intent_code(intent_code),
    _n(0),
    _position_min(0),
    _position_max(0),
    _experiment_count(0),
    _M(0),
    _N(0),
    _r(0.0),
    _Pot(0.0),
    _Pot_expectation(0.0),
    _Pot_sd(0.0),
    _z(0.0)
{ 
  Tk::Algorithm::Offset results_row;
  Tk::Algorithm::Offset offset;
  int screen_pos;
  Tk::Real screen_width;
  Position pos;
  PositionCounts pos_counts;
  PositionOffsetDistributions pos_offsets;
  
  //WDEBUG_OUT << L"Experiment count: " << experiments.size() << std::endl;
  /*
   * Compile stats
   */
  for (Db::Ids::const_iterator ex_iter(experiments.begin());
       ex_iter != experiments.end();
       ++ex_iter)
    {
      ExperimentStatistics ex;
      Count ex_n(0);
      Position ex_pos_sum(0);
      Db::Id ex_id(*ex_iter);
      
      ++_experiment_count;
      
      screen_pos = 0;
      pos = 0;
      
      screen_width = database.get_property("experiment", ex_id, "screen width").get_datum_value<long>();
      
      {
	/* get_results() needs the connection to be locked */
	Tk::Db::Connection::Mutex::Lock lock( database.get_connection()->get_mutex() );
	Tk::Db::DataSetPtr results( database.get_results(ex_id) );
	
	bool results_valid( (*results)->next() );
	if (!results_valid)
	  {
	    String msg(L"No results for experiment ");
	    msg += String(ex_id);
	    throw std::runtime_error(static_cast<std::string>(msg));
	  }
	
	for(;
	    !(*results)->isAfterLast();
	    (*results)->next())
	  {
	    offset = (*results)->getInt(1);
	    
	    screen_pos += offset;
	    pos = static_cast<Position>(screen_pos) / static_cast<Position>(screen_width);
	    
	    /* overall */
	    ++_n;
	    if (_position_max < pos)
	      _position_max = pos;
	    else if (_position_min > pos)
	      _position_min = pos;
	    
	    /* offset */
	    ++(pos_counts[pos]);
	    ++(pos_offsets[pos][offset]);
	    
	    /* experiment */
	    ++ex_n;
	    ex_pos_sum += pos;
	  }
      }
      /* complete experiment stats */
      ex._position_mean = ex_pos_sum / ex_n;
      
      if (S > 0)
	{
	  /* Record "event slots" as per Schmidt (2000) (ie, indicies of non-zero bits) */
	  long bit_count( database.get_property("experiment", ex_id, "randomness bits").get_datum_value<long>() );
	  Tk::Db::Bytes random_bytes( database.get_property("experiment", ex_id, "randomness").get_datum_value<Tk::Db::Bytes>() );
	  
	  Tk::Random::Bits random_bits( random_bytes.get_bits(bit_count) );
	  _M += bit_count;
	  SlotIndex i(1);
	  for (Tk::Random::Bits::iterator iter( random_bits.begin() );
	       iter != random_bits.end();
	       ++iter, ++i)
	    {
	      if (*iter)
		{
		  _event_slots.push_back(i);
		  ++_N;
		}
	    }
	}
      
      
      _experiments.push_back(ex);
    }
  
  /* complete overall stats */
  _offsets = OffsetStatisticsSet(offsets,
				 pos_offsets,
				 pos_counts);
  
  /*
   * Calculate mean and variance of experiments' position mean
   */
  Position square;
  Position difference;
  
  /* grand mean */
  Position ex_pos_sum(0);
  for(ExperimentStatisticsSet::iterator iter( _experiments.begin() );
      iter != _experiments.end();
      ++iter)
    ex_pos_sum += iter->_position_mean;
  
  _experiment_position_mean = ex_pos_sum / static_cast<Tk::Rational>(_experiment_count);
  
  /* variance */
  ex_pos_sum = 0;
  for(ExperimentStatisticsSet::iterator iter(_experiments.begin());
      iter != _experiments.end();
      ++iter)
    {
      difference = iter->_position_mean - _experiment_position_mean;
      square = difference * difference;
      ex_pos_sum += square;
    }
  
  _experiment_position_variance = ex_pos_sum / static_cast<Tk::Rational>(_experiment_count);
  
  
  /*
   * Calculate Pot, etc. as per Schmidt (2000)
   */
  if (_N > 0 && S > 0)
    {
      _r = exp(-static_cast<HardReal>(_N) / (static_cast<HardReal>(_M)*S));
      std::wcout << _(L"Calculating Pot for ")
		 << Tk::String(_N)
		 << _(L" events from ")
		 << Tk::String(_M)
		 << _(L" slots under intent code `")
		 << Tk::String(intent_code)
		 << L"' (S: "
		 << Tk::String(S)
		 << L"; r: " << Tk::String(_r) << L")...";
      _Pot = Pot(_r, _event_slots , _thread_profile);


#ifdef TK_POT_PROGRESS
      std::wcout << L"...";
#endif
      std::wcout << Tk::String(_Pot)
		 << std::endl;


      _Pot_expectation = Pot_expectation(_N, _M, _r);
      _Pot_sd = Pot_sd(_N, _M, _r);
      _z = (_Pot - _Pot_expectation) / _Pot_sd;
      

#ifdef TK_OPENMP_PROFILE      
      WDEBUG_OUT << L"Thread profile:" << std::endl;
      for (ThreadContextMap::iterator ctx( _thread_profile.begin() );
	   ctx != _thread_profile.end();
	   ++ctx)
	{
	  WDEBUG_OUT << L"  Context: " << ctx->first << std::endl;
	  for (ThreadCountMap::iterator cnt( ctx->second.begin() );
	       cnt != ctx->second.end();
	       ++cnt)
	    WDEBUG_OUT << L"    Thread num " << cnt->first << L": " << cnt->second << std::endl;
	}
#endif /* TK_OPENMP_PROFILE */
    }
}

AlgorithmStatistics:: AlgorithmStatistics(Db::Database &database,
					  const IntentCodeSet &intent_codes,
					  const String &algo_spec,
					  const IntentExperiments &experiments,
					  HardInteger S)
  : _algorithm_spec(algo_spec)
{
  Tk::Algorithm::Algorithm algo(algo_spec);
  _offsets = algo.get_offsets();
  
  for (IntentCodeSet::const_iterator iter(intent_codes.begin());
       iter != intent_codes.end();
       ++iter)
    {
      IntentExperiments::const_iterator intent_exs(experiments.find(*iter));
      if (intent_exs == experiments.end())
	{
	  /*
	  std::wcerr << L"Experiments for intent code '"
		     << *iter
		     << L"' not found"
		     << std::endl;
	  */
	  continue;
	}
      
      _intent_stats.push_back(Statistics(database, intent_exs->first, _offsets, intent_exs->second, S));
    }
}


TraceStatistics:: TraceStatistics (Db::Database &database,
				   unsigned trace_count,
				   const AlgorithmExperiments &experiments,
				   const IntentCodeSet &intent_codes,
				   HardInteger S)
  : _trace_count(trace_count),
    _intent_codes(intent_codes)
{
  for (AlgorithmExperiments::const_iterator iter( experiments.begin() );
       iter != experiments.end();
       ++iter)
    {
      const String &algo_spec(iter->first);
      const IntentExperiments &algo_experiments(iter->second);
      
      _algorithm_stats.push_back(AlgorithmStatistics(database, _intent_codes, algo_spec, algo_experiments, S));
    }
}

static void
gsl_error_handler (const char * reason,
		   const char * file,
		   int line,
		   int gsl_errno)
{
  std::ostringstream ostr;
  ostr << file
       << ": "
       << line
       << ": "
       << reason
       << " (GSL error code "
       << gsl_errno
       << ")";
  throw std::runtime_error(ostr.str());
}


TTest:: TTest(const Statistics &stats_a, const Statistics &stats_b)
  : _t(0.0),
    _df( stats_a._experiment_count + stats_b._experiment_count - 2 ),
    _p(0.0)
{
  // from http://www.socialresearchmethods.net/kb/stat_t.php
  /*  Tk::Rational term_a(stats_a._experiment_position_variance / stats_a._experiment_position_mean);
  Tk::Rational term_b(stats_b._experiment_position_variance / stats_b._experiment_position_mean);
  Tk::Real std_error( sqrt(static_cast<Tk::Real>(term_a + term_b)) );
  
  Tk::Rational mean_difference(stats_a._experiment_position_mean - stats_b._experiment_position_mean);
  Tk::Rational abs_mean_difference(abs(mean_difference));
  
  _t = abs_mean_difference / std_error;*/
  
  Tk::Rational mean_difference( stats_a._experiment_position_mean - stats_b._experiment_position_mean );
  
  Tk::Rational sqrt_dividend( (stats_a._experiment_count-1)*stats_a._experiment_position_variance
			  + (stats_b._experiment_count-1)*stats_b._experiment_position_variance );
  Tk::Rational sqrt_divisor(_df);
  Tk::Rational sqrt_second_term( static_cast<Tk::Rational>(1)/stats_a._experiment_count
			     + static_cast<Tk::Rational>(1)/stats_b._experiment_count );
  Tk::Real std_error( sqrt(static_cast<Tk::Real>( (sqrt_dividend/sqrt_divisor) * sqrt_second_term )) );
  
  _t = mean_difference / std_error;
  
  // from http://www.nrbook.com/a/bookcpdf/c6-4.pdf (my god what a trek)
  gsl_set_error_handler(gsl_error_handler);
  unsigned long v(_df.get_ui());
  _p = gsl_sf_beta_inc(0.5*v, 0.5, v / (v + static_cast<mpf_class>(_t*_t).get_d()));
}

Analyser:: Analyser(Db::Database &database,
		    const String &query,
		    HardInteger S)
{
  Db::Ids experiments( Query::get_experiments(&database, query) );
  
  std::wcout << Tk::String(experiments.size()) << L" experiments" << std::endl;
#ifdef _OPENMP
  std::wcout << L"Using OpenMP " << _OPENMP << std::endl;
#endif

  // Sort experiments by trace count, algorithm and intent code
  typedef std::map<unsigned, IntentCodeSet> TraceIntentCodes;
  typedef std::map<unsigned, AlgorithmExperiments> TraceExperiments;
  
  TraceIntentCodes intent_codes;
  TraceExperiments sorted_experiments;
  
  for (Db::Ids::iterator e( experiments.begin() );
       e != experiments.end();
       ++e)
    {
      Db::Id ex_id( *e );
      
      unsigned trace_count;
      String algo_spec;
      IntentCode intent_code;
      try
	{
	  trace_count = database.get_property("experiment", ex_id, "trace count").get_datum_value<long>();
	  algo_spec   = database.get_property("experiment", ex_id, "algorithm spec").get_datum_value<Tk::String>();
	  
	  Tk::StringValue ic( database.get_property("experiment", ex_id, "intent code").get_value<Tk::String>() );
	  if (ic.is_null())
	    intent_code = L"CN";
	  else
	    intent_code = ic.get_value();
	}
      catch (Tk::Db::PropertyNotSet &pns)
	{
	  continue;
	}
      
      intent_codes[trace_count].insert(intent_code);
      
      /*
      WDEBUG_OUT << L"Experiment: trace count: "
		 << trace_count
		 << L"; algo spec: "
		 << algo_spec
		 << L"; intent code: "
		 << intent_code
		 << std::endl;
      */
      sorted_experiments[trace_count][algo_spec][intent_code].push_back(ex_id);
    }

  for (TraceExperiments::iterator titer( sorted_experiments.begin() );
       titer != sorted_experiments.end();
       ++titer)
    {
      unsigned trace_count( titer->first );
      AlgorithmExperiments &algo_experiments( titer->second );
      
      _statistics.push_back(TraceStatistics(database, trace_count, algo_experiments, intent_codes[trace_count], S));
    }
}

