/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __TK_ANALYSER_H__
#define __TK_ANALYSER_H__

#include <set>

#include "gmpxx.h"

#include "libtkcommon/numerical-types.h"
#include "libtkcommon/string.h"
#include "libtkcommon/pair.h"
#include "libtkcommon/exceptions.h"
#include "libtkalgorithm/algorithm.h"
#include "libtkdb/database.h"
#include "libtkresolver/types.h"

namespace Tk {
  namespace Analysis {

    typedef Tk::String IntentCode;
    typedef std::set<IntentCode> IntentCodeSet;
    
    typedef std::list<Tk::Algorithm::Offset> OffsetList;
    typedef std::map<Tk::Algorithm::Offset, Count> OffsetDistribution;
    typedef std::map<Position, OffsetDistribution> PositionOffsetDistributions;
    typedef std::map<Position, Count> PositionCounts;
    
    typedef int                    HardInteger;
    typedef double                 HardReal;
    typedef HardInteger            SlotIndex;
    typedef std::vector<SlotIndex> Slots;
    
    class PositionFunction : public std::map<Position, Rational>
      {
      public:
	const Rational* operator()(const Position &pos) const;
        Rational calc_rss(const Probability &prob) const;
	Rational calc_mean() const;
	Rational calc_variance(const Rational &mean) const;
	Real     calc_skewness(const Rational &mean, const Rational &variance) const;
      };
    
    class OffsetStatistics
      {
      public:
	Probability              _probability;
	Count                    _position_count_max;
	PositionCounts           _position_counts;
	PositionFunction         _position_function;
	Rational                 _rss;
	Rational                 _mean;
	Rational                 _variance;
	Real                     _skewness;
      };
    typedef std::map<Tk::Algorithm::Offset, OffsetStatistics> OffsetStatisticsMap;
    
    class OffsetStatisticsSet
      {
      public:
	OffsetList               _offsets;
	OffsetStatisticsMap      _offset_stats;
	PositionCounts           _position_totals;
	
	OffsetStatisticsSet() {}
	OffsetStatisticsSet(const Tk::Algorithm::Offsets &offsets,
			    const PositionOffsetDistributions &offset_distributions,
			    const PositionCounts &position_counts);
      };
    
    class ExperimentStatistics
      {
      public:
        Position                 _position_mean;
      };
    typedef std::list<ExperimentStatistics> ExperimentStatisticsSet;
    
    class Statistics
      {
      public:

	IntentCode               _intent_code;
	
	Count                    _n;
	Position                 _position_min;
	Position                 _position_max;
	
	OffsetStatisticsSet      _offsets;
	ExperimentStatisticsSet  _experiments;
	
	Count                    _experiment_count;
	Position                 _experiment_position_mean;
	Rational                 _experiment_position_variance;
	
	typedef std::map<int, unsigned long> ThreadCountMap;
	typedef std::map<int, ThreadCountMap> ThreadContextMap;
 	ThreadContextMap         _thread_profile;
	
	HardInteger              _M;
	HardInteger              _N;
	Slots                    _event_slots;
	HardInteger              _S;
	HardReal                 _r;
	HardReal                 _Pot_expectation;
	HardReal                 _Pot_sd;
	HardReal                 _Pot;
	HardReal                 _z;
	
	static HardReal Pot(HardReal r, const Slots &event_slots , ThreadContextMap &thread_profile);
	static HardReal Pot_expectation(HardInteger N, HardInteger M, HardReal r);
	static HardReal S(HardReal r, HardInteger M);
	static HardReal G(HardReal r, HardInteger M);
	static HardReal Pot_sd(HardInteger N, HardInteger M, HardReal r);
	
	Statistics(Db::Database &database,
		   const IntentCode &intent_code,
		   const Tk::Algorithm::Offsets &offsets,
		   const Db::Ids &experiments,
		   HardInteger S);
      };
    
    typedef std::list<Statistics> StatisticsList;
    typedef std::map<IntentCode, Db::Ids> IntentExperiments;
    
    class AlgorithmStatistics
      {
      public:
	String                   _algorithm_spec;
	Tk::Algorithm::Offsets   _offsets;
	StatisticsList           _intent_stats;
	
	AlgorithmStatistics(Db::Database &database,
			    const IntentCodeSet &intent_codes,
			    const String &algo_spec,
			    const IntentExperiments &experiments,
			    HardInteger S);
    };
    typedef std::list<AlgorithmStatistics> AlgorithmStatisticsList;
    typedef std::map<String, IntentExperiments> AlgorithmExperiments;
    
    class TraceStatistics
      {
      public:
	
	unsigned                 _trace_count;
	IntentCodeSet            _intent_codes;
	AlgorithmStatisticsList  _algorithm_stats;
	
	TraceStatistics(Db::Database &database,
			unsigned trace_count,
			const AlgorithmExperiments &experiments,
			const IntentCodeSet &intent_codes,
			HardInteger S);
      };
    typedef std::list<TraceStatistics> TraceStatisticsList;
    
    class TTest
      {
      private:
      public:
	Real     _t;
	Integer  _df;
	Real     _p;
	
	TTest(const Statistics &stats_a, const Statistics &stats_b);
      };
    
    class Analyser
      {
      public:
	TraceStatisticsList _statistics;
	
	Analyser(Db::Database &database, const String &query_file, HardInteger S = 1);
      };

    
    class NotEnoughResults : std::runtime_error
      {
      public:
	NotEnoughResults()
	  : std::runtime_error("not enough results") {}
      };

    
  }; /* namespace Analysis */
  
  double gamma(double x);
}; /* namespace Tk */


#endif /* __TK_ANALYSER_H__ */

