/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <wchar.h>
#include <errno.h>

#include "mbs.h"

using namespace Tk;

#define DEBUG 1

#ifdef DEBUG
#include <cstdio>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>

static void
record_buf_size(const char *conv_func, int buf_size, int used)
{
  int err;
  int fd;
  FILE *fs;
  std::string fn;
  
  // construct file name
  fn = getenv("HOME");
  fn += "/libtkcommon-mbs-conv-buf-sizes.log";
  
  // open file
  fd = open(fn.c_str(), O_WRONLY|O_APPEND|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if (fd == -1)
    throw UnixError("Could not open buf size record file", errno);
  
  // lock file
  err = flock(fd, LOCK_EX);
  if (err == -1)
    throw UnixError("Could not lock buf size record file", errno);
  
  // create stream
  fs = fdopen(fd, "a");
  if (fs == NULL)
    throw UnixError("Could not open buf size record file stream from file descriptor", errno);
  
  // write to stream
  err = fprintf(fs, "%s %d %d\n", conv_func, buf_size, used);
  if (err < 0)
    throw UnixError("Could not write to buf size record file stream", errno);
  
  // close file/stream
  err = fclose(fs);
  if (err == EOF)
    throw UnixError("Could not close buf size record file stream", errno);
}
#endif /* DEBUG */

wchar_t *
Tk:: mbstowcs_alloc(const char *string)
{
  mbstate_t state;
  const char *str = string;
  size_t buf_size = strlen(string) + 1;
  wchar_t *buf = static_cast<wchar_t *>( malloc (buf_size * sizeof(wchar_t)) );
  wchar_t *conversion_start = buf;
  size_t buf_done = 0;
  size_t converted;

  memset(&state, '\0', sizeof(state));
      
  while (1)
    {
      converted = mbsrtowcs(conversion_start, &str, buf_size - buf_done, &state);
	  
      if (converted == (size_t)(-1))
	throw UnixError("Could not convert multi-byte string to wide-character string: ", errno);
	  
      if (str == NULL)
	break;
	  
      buf_done += converted;
      buf_size += 8;
      buf = static_cast<wchar_t *>( realloc(buf, buf_size * sizeof(wchar_t)) );
      conversion_start = buf + buf_done;
    }
      
#ifdef DEBUG
  record_buf_size("mbstowcs_alloc", buf_size, wcslen(buf) + 1);
#endif

  return buf;
}

char *
Tk:: wcstombs_alloc(const wchar_t *string)
{
  mbstate_t state;
  const wchar_t *str = string;
  size_t buf_size = wcslen(string) + 1;
  char *buf = static_cast<char *>( malloc (buf_size * sizeof(char)) );
  char *conversion_start = buf;
  size_t buf_done = 0;
  size_t converted;
  
  memset(&state, '\0', sizeof(state));
  
  while (1)
    {
      converted = wcsrtombs(conversion_start, &str, buf_size - buf_done, &state);
	  
      if (converted == (size_t)(-1))
	throw UnixError("Could not convert wide-character string to multi-byte string: ", errno);
	  
      if (str == NULL)
	break;
	  
      buf_done += converted;
      buf_size += 8;
      buf = static_cast<char *>( realloc(buf, buf_size * sizeof(char)) );
      conversion_start = buf + buf_done;
    }

#ifdef DEBUG  
  record_buf_size("wcstombs_alloc", buf_size, strlen(buf) + 1);
#endif

  return buf;
}
