/* -*- c++ -*-
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Should you need to contact me, the author, you can do so by
 * e-mail - mail your message to Viktor T. Toth at <vttoth@vttoth.com>.
 *
 * Updated 2007 by Robert Ham for tkresolver
 *
 * Copyright (C) 2007 Robert Ham <rah@bash.sh> for tkresolver
 * Copyright (C) Viktor T. Toth at <vttoth@vttoth.com>
 *
 */

#include <sstream>

#include "real.h"

using namespace Tk::MP;

Real::Pi Tk::MP::Real::pi(1024);
Real::E  Tk::MP::Real::e(1024);

Real::Pi:: Pi(unsigned long prec)
{
  set_prec(prec-1);
  set(prec);
}

void
Real::Pi:: set(unsigned long prec)
{
  if (get_prec() == prec)
    return;
  
  set_prec(prec);
  
  Number x( 6 - 4 * sqrt(Number(2)) );
  Number y( sqrt(Number(2)) - 1 );
  Number two(2);
  
  while (true)
    {
      two *= 4;
      Number t( sqrt(sqrt(1 - y*y*y*y)) );
      Number y_( (1 - t) / (1 + t) );
      t = 1 + y_;
      Number x_( (t*t*t*t * x) - two * y_ * (1 + y_ + y_*y_) );
      
      if (x_ == x)
	break;
      
      x = x_;
      y = y_;
    }
  
  *(static_cast<Number *>(this)) = static_cast<Number>(1 / x);
}

Real::E:: E(unsigned long prec)
{
  set_prec(prec-1);
  set(prec);
}

void
Real::E:: set(unsigned long prec)
{
  if (get_prec() == prec)
    return;
  
  set_prec(prec);
  
  *(static_cast<Number *>(this)) = 1.0;
  Number s(1, prec);
  
  for (int i = 1; ; ++i)
    {
      s /= i;
      Number t = *this + s;
      
      if (t == *this)
	break;
      
      *(static_cast<Number *>(this)) = t;
    }
}


Real::Number
Real:: pow(const Number &op1, unsigned long op2)
{
  Number r;
  mpf_pow_ui(r.get_mpf_t(), op1.get_mpf_t(), op2);
  return r;
}

Real::Number
Real:: pow(const Number &op1, const Number &op2)
{
  return exp(op2 * log(op1));
}

Real::Number
Real:: log(const Number &op_)
{
  if (op_ <= 0.0)
    {
      std::ostringstream ostr;
      ostr << "log("
	   << op_
	   << ") is undefined"
	;
      throw std::invalid_argument(ostr.str());
    }
  
  /*
  if (op < 0) return mpdouble(0);

  mpdouble r;
  while (op > 1.5)
    {
      r += 1;
      op /= mpdouble::e();
    }
  while (op < .5)
    {
      r -= 1;
      op *= mpdouble::e();
    }

  op = op - 1;
  mpdouble t = r;
  mpdouble o = 1;
  int s = -1;
  for (int i = 1; ; i++)
    {
      o *= op;
      s = -s;
      t += o / (s*i);
      if (t == r) break;
      r = t;
    }
  return r;
  */


  Number op(op_);
  Number r;
  while (op > 1.5)
    {
      r += 1;
      op /= Real::e;
    }
  while (op < .5)
    {
      r -= 1;
      op *= Real::e;
    }

  op -= 1;
  Number t(r);
  Number o(1);
  int s = -1;
  for (int i = 1; ; ++i)
    {
      o *= op;
      s = -s;
      t += o / (s*i);
      
      if (t == r)
	break;
      
      r = t;
    }
  
  return r;
}

Real::Number
Real:: sin(const Number &x_)
{
  Number t( floor(x_ / (2*Real::pi)) );
  Number x(x_ - t * 2 * Real::pi);
  if (x > Real::pi)
    x -= 2 * Real::pi;
  
  t = x;
  Number p(x);
  Number r(x);
  Number xx(x*x);
  int i = 1;
  while (true)
    {
      p *= -xx / ++i / ++i;
      t = r + p;
      
      if (t == r)
	break;
      
      r = t;
    }
  
  return r;
}

Real::Number
Real:: cos(const Number &x)
{
  return sin(Real::pi / 2 - x);
}

Real::Number
Real:: exp(const Number &x_)
{
  if (x_ == 0)
    return Number(1);
  
  if (x_ < 0)
    return 1/exp(-x_);
  
  Number x(x_);
  Number t( floor(x) );
  x -= t;
  Number p( pow(e, t.get_ui()) );
  
  Number s(1);
  Number r(1);
  for (int i = 1; ; i++)
    {
      s *= x / i;
      t = r + s;
      
      if (t == r)
	break;
      
      r = t;
    }
  
  return r * p;
}

Real::Number
Real:: atan2(const Number &y, const Number &x)
{
  Number r;
  if (abs(y) > abs(x))
    r = Real::pi / 2 - atan2(x, y);
  else
    {
      if (abs(y) == abs(x))
	r = Real::pi / 4;
      else
	if (abs(y) == 0)
	  r = 0;
	else
	  {
	    r = abs(y/x);
	    Number yx(r*r);
	    Number s(r);
	    
	    for (int i = 3; ; i += 2)
	    {
	      s = -s * yx;
	      Number t(r + s/i);
	      
	      if (t == r)
		break;
	      
	      r = t;
	    }
	}
      
      if (y < 0)
	r = -r;
      
      if (x < 0)
	r = Real::pi - r;
    }
  
  if (r < -Real::pi)
    r += 2 * Real::pi;
  
  if (r > Real::pi)
    r -= 2 * Real::pi;
  
  return r;
}

