/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdexcept>
#include <cmath>

#include "libtkcommon/debug.h"
#include "libtkcommon/maths.h"
#include "libtkcommon/intl.h"
#include "libtkrandom/decision_tree.h"
#include "libtkrandom/randomness.h"
#include "libtkrandom/retriever_registry.h"

using namespace Tk;
using namespace Random;



DecisionTree:: DecisionTree(unsigned choice_count, unsigned bit_multiplier)
{
  if (bit_multiplier % 2 == 0)
    throw std::invalid_argument(_("Bit multiplier must be odd"));
  
  // recursively generate tree
  _root = new Node(choice_count, bit_multiplier);
}

DecisionTree:: ~DecisionTree ()
{
  delete _root;
}

unsigned
DecisionTree:: get_depth(unsigned choice_count)
{
  return find_next_power_of_two_power(choice_count);
}

unsigned
DecisionTree:: tree_leaf_count(unsigned choice_count)
{
  return 1U << get_depth(choice_count);
}

double
DecisionTree:: expected_tries(unsigned choice_count)
{
  unsigned leaf_count(tree_leaf_count(choice_count));
  
  double expected_tries = 1.0;
  if (leaf_count != choice_count)
    {
      // approximating an infinite function
      const unsigned iterations = 15;
      double retry_rate( (leaf_count - choice_count) / static_cast<double>(leaf_count) );
      double retries(0.0);

      for (unsigned i = 1; i < iterations; ++i)
	retries += pow(retry_rate, i);

      expected_tries += retries;
    }

  return expected_tries;
}

bool
DecisionTree:: deterministic(unsigned choice_count)
{
  return tree_leaf_count(choice_count) == choice_count;
}

unsigned
DecisionTree:: decide(Retriever &retriever, Randomness &randomness, Effector &effector)
{
  DecisionTree::Collector collector(_root, randomness, effector);
  retriever.retrieve(collector);
  return collector.get_retries();
}

DecisionTree::Collector:: Collector(const Node *root, Randomness &randomness, Effector &effector)
  : _root(root),
    _randomness(randomness),
    _effector(effector),
    _retries(0),
    _decision_state(0)
{
}

DecisionTree::Collector:: ~Collector()
{
  if (_decision_state)
    delete _decision_state;
}

bool
DecisionTree::Collector:: collect(Buffer *buffer)
{
  _randomness.add_buffer(*buffer);
  delete buffer;
  
  DecisionState state(_root);
  if (_decision_state)
    {
      state = *_decision_state;
      delete _decision_state;
      _decision_state = 0;
    }
  
  try
    {
      if (!_effector.affect(Node::choose(*_root, _randomness, _retries, state)))
	return false;
      
      while (_effector.affect(_root->choose(*_root, _randomness, _retries)));
    }
  catch (PartialTraversal pt)
    {
      _decision_state = pt.get_decision_state();
      /*
      WDEBUG_OUT << L"Partial traversal, bit count; "
		 << _decision_state->get_bit_count()
		 << L"; bit index: "
		 << _decision_state->get_bit_index()
		 << std::endl;
      */
      return true;
    }
  
  return false;
}

unsigned
DecisionTree::Collector:: get_retries() const
{
  return _retries;
}

DecisionTree::DecisionState:: DecisionState(const Node *node, unsigned bit_count, unsigned bit_index)
  : _node(node),
    _bit_count(bit_count),
    _bit_index(bit_index)
{}

const DecisionTree::Node *
DecisionTree::DecisionState:: get_node() const
{
  return _node;
}

unsigned
DecisionTree::DecisionState:: get_bit_count() const
{
  return _bit_count;
}

unsigned
DecisionTree::DecisionState:: get_bit_index() const
{
  return _bit_index;
}

DecisionTree::PartialTraversal:: PartialTraversal(DecisionState *state)
  : _state(state)
{} 

DecisionTree::DecisionState *
DecisionTree::PartialTraversal:: get_decision_state() const
{
  return _state;
}

DecisionTree::Node:: Node(unsigned choice_count, unsigned bit_multiplier)
{
  Choice choice(0);
  init(choice_count, bit_multiplier, DecisionTree::get_depth(choice_count), choice);
}

DecisionTree::Node:: Node(unsigned choice_count, unsigned bit_multiplier, unsigned depth, Choice &choice)
{
  init(choice_count, bit_multiplier, depth, choice);
}

void
DecisionTree::Node:: init(unsigned choice_count, unsigned bit_multiplier, unsigned depth, Choice &choice)
{
  /*
  WDEBUG_OUT << L"Initialising node at depth "
	     << depth
	     << L"; choice count: "
	     << choice_count
	     << L"; current choice: "
	     << choice
	     << std::endl;
  */
  _bit_multiplier = bit_multiplier;

  if (depth == 0)
    {
      _choice = new Choice(choice++);
      _children[0] = _children[1] = 0;
    }
  else
    {
      _choice = 0;
      _children[0] = new Node(choice_count, bit_multiplier, depth - 1, choice);
      if (choice < choice_count)
	_children[1] = new Node(choice_count, bit_multiplier, depth - 1, choice);
      else
	_children[1] = 0;
    }
}



DecisionTree::Node:: ~Node ()
{
  if (_choice)
    delete _choice;
  if (_children[0])
    delete _children[0];
  if (_children[1])
    delete _children[1];
}

bool
DecisionTree::Node:: decide(Randomness& rand, unsigned bit_count, unsigned bit_index) const
{
  bool bit;
  for (; bit_index < _bit_multiplier; ++bit_index)
    {
      try
	{
	  bit = rand.get_bit();
	}
      catch (Randomness::OutOfBits)
	{
	  //WDEBUG_OUT << L"Out of bits while deciding" << std::endl;
	  throw PartialTraversal( new DecisionState(this, bit_count, bit_index) );
	}
      bit_count += bit;
    }
  
  return bit_count > ((_bit_multiplier - 1) / 2);
}

Choice
DecisionTree::Node:: choose(const Node &root, Randomness &rand, unsigned &retries, unsigned bit_count, unsigned bit_index) const
{
  if (_choice)
    return *_choice;
  
  if (decide(rand, bit_count, bit_index))
    {
      return _children[0]->choose(root, rand, retries);
    }
  else
    {
      if (!_children[1])
	{
	  ++retries;
	  return root.choose(root, rand, retries);
	}
      else
	return _children[1]->choose(root, rand, retries);
    }
}

Choice
DecisionTree::Node:: choose(const Node &root, Randomness &rand, unsigned &retries, const DecisionState &decision_state)
{
  return decision_state.get_node()->choose(root, rand, retries, decision_state.get_bit_count(), decision_state.get_bit_index());
}
