/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdexcept>
#include <iomanip>

#include <glibmm.h>
#include <quantis.h>

#include "libtkrandom/quantis.h"

using namespace Tk::Random;

const unsigned short Quantis::BLOCK_SIZE(32);
int                  Quantis::_card_no(-1);
Quantis::Mutex       Quantis::_lock = GLIBMM_STATIC_MUTEX_INIT;

void
Quantis:: init()
{
  int ver( quantisLibVersion() );
  double ver_decode( ver / 10.0 );
  std::wcout << L"Quantis library version "
	     << std::setprecision(1)
	     << std::fixed
	     << ver_decode
	     << std::endl;
  
  int card_count( quantisCount() );
  if (card_count == 0)
    throw std::runtime_error("Error counting Quantis devices (possibly no devices)");
    
  ver = quantisDriverVersion();
  if (ver == 0)
    throw std::runtime_error("Error retrieving driver version");
  ver_decode = ver / 10.0;
  std::wcout << L"Quantis driver version "
	     << ver_decode
	     << std::endl;
  
  // FIXME: have some method of specifying card number
  // More generally need a method of specifying retriever parameters
  _card_no = 1;
  std::wcout << L"Using Quantis card #"
	     << _card_no
	     << std::endl;
  
  
  ver = quantisBoardVersion(_card_no);
  if (ver == 0)
    throw std::runtime_error("Error retrieving Quantis board version");
  
  static const int y_mask = 0xff000000;
  static const int m_mask = 0x00ff0000;
  static const int d_mask = 0x0000ff00;
  static const int r_mask = 0x000000ff;
  
  int year  = ver & y_mask >> 24;
  int month = ver & m_mask >> 16;
  int day   = ver & d_mask >>  8;
  int rev   = ver & r_mask >>  0;
  
  std::wcout << L"Quantis board version: "
	     << std::setfill(L'0')
	     << std::setw(2) << day
	     << L"/"
	     << std::setw(2) << month
	     << L"/"
	     << std::setw(2) << year
	     << " revision "
	     << rev
	     << std::endl;
  
  
  int ret( quantisBoardReset(_card_no) );
  if (ret < 0)
    throw std::runtime_error("Error resetting Quantis device");
}

Quantis:: Quantis(unsigned count)
  : Retriever(count)
{
  Glib::Mutex::Lock lock( _lock );
  
  if (_card_no == -1)
    init();
}

void
Quantis:: retrieve(Collector &collector)
{ 
  Buffer *buf;
  char raw_buf[BLOCK_SIZE];
  
  Glib::Mutex::Lock lock( _lock );
  do
    {
      int ret( quantisRead(_card_no, raw_buf, BLOCK_SIZE) );
      if (ret == -1)
	throw std::runtime_error("Error reading from Quantis device");
      
      // On success, the exact number of bytes requested is returned
      
      buf = new Buffer;
      for (int i = 0; i < BLOCK_SIZE; ++i)
	buf->push_back(raw_buf[i]);
    }
  while (collector.collect(buf));
}

bool
QuantisType:: is_true_randomness()
{
  return true;
}

unsigned
QuantisType:: get_preferred_block_size()
{
  return Quantis::BLOCK_SIZE;
}

Retriever *
QuantisType:: new_retriever(unsigned count)
{
  return new Quantis(count);
}
